//  RW Dolldrop
// ==============
// External File Transfers Library
// -Part of the RW Foundations suite of libraries

// Filename: ./include/dollhouse/net/dolldrop.hpp
// -Declarations of the baseline RW Dolldrop sub-library

#pragma once

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

//////////////////
//  'Net Work :
//

// DESIGN:
// -As a simple, basic control mechanism for starters we'll use some
// protected/private classes and a few functions from here within Dolldrop.
// Later, we can add more access restrictions and some further data validations;
// improving things bit-by-bit as we go along, using just a few of the basic &
// straightforward language features readily available to us in the standard.

// DESIGN:
// - add work-stealing queues of the Dolldrops/Mailboxes ?
// - use S.Parent's libdispatch-based async
// - this is actually a general need, so add it to the RW Foundations instead

// DESIGN:
// void push_drop() {}
// void yank_drop() {}

// DESIGN: TODO:
// return do_atomiconn(
//   lockout();
//     connect();
//       do_req(uri);
//     disconnect();
//   unlockout();
// );

// DESIGN: during this prototyping phase, code-envy is acceptable. however, for
// the production code abstract common behaviours into modular functions.
// -DRY principle

namespace rw::dollhouse::net {

//------------------------------------------------------------------------------
/**
 */
class Dolldrop_b : public RW_slave {
 public:
  Dolldrop_b(const Hip& tag, std::string const& handle);  // tag-parm'd ctor

 protected:
  //-------
  inline auto ask_drop(Hip const& tag, std::string const& uri)
      -> std::ifstream {
    if (tag == master())                    // 1st confirm dolldrop's Master
      return mailstop_.try_drop(t(), uri);  // then redirect to mailstop
    else
      only_master_can(master(), tag, t(), "Dolldrop_b::ask_drop()");

    ///
    return ifstream{};  // shouldn't reach here
  }

  //-------
  inline auto ask_drop_hdr(Hip const& tag, std::string const& uri)
      -> std::ostringstream {
    if (tag == master())                        // 1st confirm dolldrop's Master
      return mailstop_.try_drop_hdr(t(), uri);  // then redirect to mailstop
    else
      only_master_can(master(), tag, t(), "Dolldrop_b::ask_drop_hdr()");

    ///
    return ostringstream{};  // shouldn't reach here
  }

  //-------
  inline auto ask_hdr_only(Hip const& tag, std::string const& uri)
      -> std::ostringstream {
    if (tag == master())                        // 1st confirm dolldrop's Master
      return mailstop_.try_hdr_only(t(), uri);  // then redirect to mailstop
    else
      only_master_can(master(), tag, t(), "Dolldrop_b::ask_hdr_only()");

    ///
    return ostringstream{};  // shouldn't reach here
  }

 private:
  Dolldrop_b() = delete;  // no default ctor

  //----------------------------------------------------------------------------
  /**
   */
  class Mailstop {
   public:
    Mailstop(Hip const& tag)
        : owner_{tag}, dollnet_{owner_, "dolldrop_dollnet"} {}

    inline auto try_drop(Hip const& tag, std::string const& uri)
        -> std::ifstream {
      return dollnet_.dl_body_only(tag, uri);  // TEMP: deliver directly for now
    }

    inline auto try_drop_hdr(Hip const& tag, std::string const& uri)
        -> std::ostringstream {
      return dollnet_.dl_w_header(tag, uri);  // TEMP: deliver directly for now
    }

    inline auto try_hdr_only(Hip const& tag, std::string const& uri)
        -> std::ostringstream {
      return dollnet_.dl_hdr_only(tag, uri);  // TEMP: deliver directly for now
    }

   private:
    Hip owner_;

    class Inbox {
      //
     private:
      string data_;
    } inbox_;

    class Outbox {
      //
     private:
      string data_;
    } outbox_;

    //--------------------------------------------------------------------------
    /**
     */
    class Dollnet : public RW_slave {
     public:
      Dollnet(Hip const& tag, std::string const& handle);  // tag-parm'd ctor

      /** Attempts an external network connection with, and download transfer
       * from, @a uri. If successful, transfer's body stream for @a uri is
       * saved out to disk file ( ./dolldrops/anons_file_nnn ).
       * @param[in] tag the caller's @c Hip
       * @param[in] uri the URI to attempt a download connection with
       * @return the disk file's handle
       * @note: internally catches a specialization of @c std::runtime_error
       * on @b curl errors (but doesn't re-throw, currently)
       */
      auto dl_body_only(Hip const&         tag,  // guarded; for Master only
                        std::string const& uri) -> std::ifstream;

      /** Attempts an external network connection with, and download transfer
       * from, @a uri. If successful, transfer's body stream for @a uri is
       * saved out to disk file ( ./dolldrops/anons_file_nnn ).
       * @param[in] tag the caller's @c Hip
       * @param[in] uri the URI to attempt a download connection with
       * @return header stream data for @a uri, if present
       * @note: internally catches a specialization of @c std::runtime_error
       * on @b curl errors (but doesn't re-throw, currently)
       */
      auto dl_w_header(Hip const&         tag,  // guarded; for Master only
                       std::string const& uri) -> std::ostringstream;

      /** Attempts an external network connection with, and header-only download
       * transfer from, @a uri
       * @param[in] tag the caller's @c Hip
       * @param[in] uri the URI to attempt a download connection with
       * @return header stream data for @a uri, if present
       * @note: internally catches a specialization of @c std::runtime_error
       * on @b curl errors (but doesn't re-throw, currently)
       */
      auto dl_hdr_only(Hip const&         tag,  // guarded; for Master only
                       std::string const& uri) -> std::ostringstream;

     private:
      Dollnet() = delete;  // no default ctor

    } dollnet_;

  } mailstop_{t()};  // construct this member w/ its enclosing Dolldrop as owner

};  // class Dolldrop_b

//------------------------------------------------------------------------------
/**
 */
class Dolldrop : public Dolldrop_b {
 public:
  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file
   * @param[in] tag the caller's @c Hip
   * @param[in] uri the URI to attempt a download connection with
   * @return the disk file's handle
   */
  auto pull(Hip const& tag, std::string const& uri) -> std::ifstream {
    stamp(tag, "pull()");

    prt_tag(tag, "asked to pull a dolldrop");
    cout << "  -requested pull address :  '" << uri << "'\n\n";

    // DESIGN: basic validation for Type of the caller 'tag'
    //  -we can implement an 'any_of()' algorithm later if needed. but for
    //  now, Type::Bumpmaster and Type::Curator are the only valid RW types
    //  for dolldrop info requests.
    //
    // if (tag.typ == Type::Bumpmaster || tag.typ == Type::Curator)
    if (tag.typ == Type::Bumpmaster)
      return ask_drop(tag, uri);
    else
      error("Invalid type for dolldrops", t(), "Dolldrop::pull()");

    ///
    return ifstream{};  // shouldn't reach here
  }

  /** Attempts an external network connection with, and download transfer from,
   * @a uri. If successful, transfer's body stream for @a uri is saved out to
   * disk file
   * @param[in] tag the caller's @c Hip
   * @param[in] uri the URI to attempt a download connection with
   * @return header stream data for @a uri, if present
   */
  auto pull_w_hdr(Hip const& tag, std::string const& uri)
      -> std::ostringstream {
    stamp(tag, "pull_w_hdr()");

    prt_tag(tag, "asked to pull a dolldrop w/ hdr");
    cout << "  -requested pull address :  '" << uri << "'\n\n";

    // if (tag.typ == Type::Bumpmaster || tag.typ == Type::Curator)
    if (tag.typ == Type::Bumpmaster)
      return ask_drop_hdr(tag, uri);
    else
      error("Invalid type for dolldrops", t(), "Dolldrop::pull_w_hdr()");

    ///
    return ostringstream{};  // shouldn't reach here
  }

  /** Attempts an external network connection with, and header-only download
   * transfer from, @a uri
   * @param[in] tag the caller's @c Hip
   * @param[in] uri the URI to attempt a download connection with
   * @return the header-only response
   */
  auto pull_hdr(Hip const& tag, std::string const& uri) -> std::ostringstream {
    stamp(tag, "pull_hdr()");

    prt_tag(tag, "asked to pull a dolldrop header");
    cout << "  -requested pull address :  '" << uri << "'\n\n";

    // if (tag.typ == Type::Bumpmaster || tag.typ == Type::Curator)
    if (tag.typ == Type::Bumpmaster)
      return ask_hdr_only(tag, uri);
    else
      error("Invalid type for dolldrops", t(), "Dolldrop::pull_hdr()");

    ///
    return ostringstream{};  // shouldn't reach here
  }

 private:
  Dolldrop() = delete;  // no default ctor

};  // class Dolldrop

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "______"

// rw_dollhouse_lib.v220115
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
